import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;

import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

/** An instance is a pen in a panel within a JFrame. <br>
 * It can be used to draw rectangles and circles, fill<br>
 * rectangles and circles, write strings, etc.
 *
 * @Author - David Gries */
public abstract class Pen {
	/** The frame on which the pen draws. */
	private static JFrame jframe;

	/** The panel in the JFrame. */
	private static JPanel panel;

	/** The graphics associated with the panel. */
	static Graphics graphics;

	/** The Controls panel in this pen. */
	Controls controls;

	/** The panel on which the pen draws is of size (width, height). */
	public static final int width= 840;

	/** The panel on which the pen draws is of size (width, height). */
	public static final int height= 800;

	/** The pen is at point (x, y). */
	private double x, y;

	/** the pen color. */
	private Color penColor= Color.black;

	/** Constructor: a black pen showing the beginnings of a maze. <br>
	 * The pen is down. <br>
	 * The maze size is (size.x, size.y), in (2..30, 2..30).<br>
	 * The start point is (st.x, st.y). <br>
	 * The end point is (en.x, en.y). <br>
	 * Precondition: The start and end points are in the maze boundary. <br>
	 * ............. The start and end points are different. */
	public Pen(Point size, Point st, Point en) {
		if (jframe == null) {
			jframe= new JFrame("Pen window");
			panel= new JPanel();
			panel.setPreferredSize(new Dimension(width, height));
			panel.setBackground(Color.white);
			jframe.getContentPane().add(panel, BorderLayout.CENTER);

			controls= new Controls(this, size, st, en);
			jframe.getContentPane().add(controls, BorderLayout.NORTH);

			jframe.getContentPane().add(new JLabel("   "), BorderLayout.WEST);

			jframe.pack();
			jframe.setVisible(true);
			graphics= panel.getGraphics();
			pause(100);
			clear();
			pause(100);
			jframe.setResizable(false);
		}
	}

	/** Process a click of new Maze button in the Control panel. <br>
	 * Precondition: no errors in the controls. Overridden in MazeGeneration. <br>
	 * Parameters are: width and height of maze, start point (sX, sY), and end point (eX, eY) */
	public void processNewMaze(int wdth, int hght, int sX, int sY, int eX, int eY) {}

	/** Set new seed to n. */
	public abstract void newSeed(int n);

	/** Generate next node of the maze. */
	public abstract void addOneNode();

	/** Complete the maze. */
	public abstract void completeTheMaze();

	/** Set a field to indicate that parent arrows are to be shown if b is true. */
	public abstract void processParentButtonClick(boolean b);

	/** = the Color value for c. The values are as follows: <br>
	 * 1 black, 2 blue, 3 cyan, 4 dark gray, 5 gray, 6 green, 7 light gray, <br>
	 * 8 magenta, 9 orange, 10 pink, 11 red, 12 white, 13 yellow, <br>
	 * any other integer black. */
	public static Color tColor(int c) {
		if (c == 1) return Color.black;
		if (c == 2) return Color.blue;
		if (c == 3) return Color.cyan;
		if (c == 4) return Color.darkGray;
		if (c == 5) return Color.gray;
		if (c == 6) return Color.green;
		if (c == 7) return Color.lightGray;
		if (c == 8) return Color.magenta;
		if (c == 9) return Color.orange;
		if (c == 10) return Color.pink;
		if (c == 11) return Color.red;
		if (c == 12) return Color.white;
		if (c == 13) return Color.yellow;
		return Color.black;
	}

	/** Set panel size as big as possible in the current window. */
	public void setPanelSize() {
		Dimension d= new Dimension(width, height);
		panel.setPreferredSize(d);
		graphics= panel.getGraphics();
		jframe.pack();
		clear();
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		pause(100);
	}

	/** = x-coordinate of the pen. */
	public double getX() {
		return x;
	}

	/** = y-coordinate of the panel. */
	public double getY() {
		return y;
	}

	/** = width of panel. */
	public int getWidth() {
		return width;
	}

	/** = height of the panel. */
	public int getHeight() {
		return height;
	}

	/** Set the color of the pen to c. */
	public void setColor(Color c) {
		penColor= c;
	}

	/** = the current color. */
	public Color getColor() {
		return penColor;
	}

	/** Move the pen to (x,y), without drawing. */
	public void moveTo(double x, double y) {
		Color save= graphics.getColor();
		graphics.setColor(penColor);
		this.x= x;
		this.y= y;
		graphics.setColor(save);
	}

	/** Draw a circle with center at current pen position of diameter d. */
	public void drawCircle(double d) {
		Color save= graphics.getColor();
		graphics.setColor(penColor);
		int id= (int) Math.round(d);
		graphics.drawOval((int) Math.round(x - d / 2), (int) Math.round(y - d / 2), id, id);
		graphics.setColor(save);
	}

	/** Fill a circle with center at current pen position of diameter d. */
	public void fillCircle(double d) {
		Color save= graphics.getColor();
		graphics.setColor(penColor);
		int id= (int) Math.round(d);
		graphics.fillOval((int) Math.round(x - d / 2), (int) Math.round(y - d / 2), id, id);
		graphics.setColor(save);
	}

	/** Draw a rectangle with center at current pen position of width b, height h. */
	public void drawRectangle(double w, double h) {
		Color save= graphics.getColor();
		graphics.setColor(penColor);
		graphics.drawRect((int) Math.round(x - w / 2), (int) Math.round(y - h / 2),
			(int) Math.round(w), (int) Math.round(h));
		graphics.setColor(save);
	}

	/** Fill a rectangle with center at current pen position of width w, height h. */
	public void fillRectangle(double w, double h) {
		Color save= graphics.getColor();
		graphics.setColor(penColor);
		graphics.fillRect((int) Math.round(x - w / 2), (int) Math.round(y - h / 2),
			(int) Math.round(w), (int) Math.round(h));
		graphics.setColor(save);
	}

	/** Put s at current pen's position in current color. */
	public void drawString(String s) {
		graphics.drawString(s, (int) x, (int) y);
	}

	/** Clear the screen (make it all white). */
	public void clear() {
		Color c= graphics.getColor();
		graphics.setColor(Color.white);
		graphics.fillRect(0, 0, width, height);
		graphics.setColor(c);
	}

	/** Pause the program for msec milliseconds. */
	public void pause(int msec) {
		try {
			Thread.sleep(msec);
		} catch (InterruptedException e) {}
	}

}